/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/******************************************************************************
 * cell programming example
 * Copyright IBM Corp. 2007 All Rights Reserved 
 * 
 * sphereMC_ppu.c:  PPE program 
 *
 * This PPE programm calculates the volume of a n-dimensional sphere using Monte Carlo simulation
 * The problem is broken into two parts and each part is solved on multiple SPEs in parallel:
 * part_1: creating a certain amount of random numbers and store them in main memory
 * part_2: use created RNs to run the Monte Carlos simulation on multiple SPEs
 *
 * Type of problem: pipeline approach with synchronization over PPE 
 *
 * This PPU example will demonstrate:  
 * - how to set up a SPE task (1)
 * - how to do static aligned memory allocation (2)
 * - how to do dynamic aligned memory allocation (3)
 * - how to set up a SPE control block (4)
 * - how to create spe thread (5)
 * - how to use maiboxes (6)
 * - how to restart a SPEs immediately with a new task (7)
 *   
 *****************************************************************************/

#include <string.h>
#include <stdio.h>
#include <libspe2.h>
#include <pthread.h>
#include <libmisc.h>

// Note: These two are order dependent to ensure the Sobol control block is defined.
#include <mc_rand_sb.h>
#include <../myType.h>
#include "../../sobol_init_30_40.h"


#define NUMSPUS 6

int get_args(int argc, char** argv, UINT32 *u32Dim, UINT32 *u32Runs, UINT32 *u32Iter, UINT8 *u8Report)
{
	int i;

	// argument checking and extraction
	if ( (argc!=7) && (argc!=8)){
      printf("\nUSAGE: %s -d <dimensins> -r <runs> -i <iterations> [-l] [-j] \n", argv[0]);
      printf(" If -l option is set a long report is printed.");
      printf(" If -j option is set a reduced report is printed.");
      printf(" If neither -l nor -j is set only mean, sigma and delta in percent of the true value is printed");
      printf(" note: dimension x runs < 7 000  and dimension x runs x iterations < 6 250 000\n");
      printf("if not program terminates \n");
      printf("Example: %s -d 4 -r 100 -i 200 \n", argv[0]);
      exit (0);
	}

	i=atoi(argv[2]);
	if (i < 0) 
	{
		printf("error: dimension must be > 0!\n");
		return 1;
	}
  
	i=atoi(argv[4]);
	if (i < 1) 
	{
		printf("error: dimension must be > 0!\n");
		return 1;
	}

	i=atoi(argv[6]);
	if (i < 1)
	{
		printf("error: iterations must be > 0!\n");
		return 1;
	}

	*u8Report = 0;
    // Start at i = 1 to skip the command name
	for (i = 1; i < argc; i++) {

		// Check for a switch (leading "-")
		if (argv[i][0] == '-') {

			// Use the next character to decide what to do
            switch (argv[i][1]) {
				case 'd':       *u32Dim = atoi(argv[++i]); break;
				case 'r':       *u32Runs = atoi(argv[++i]); break;
				case 'i':       *u32Iter = atoi(argv[++i]); break;
				case 'l':       *u8Report = 1; ++i; break;
                case 'j':       *u8Report = 2; ++i; break;
				default:        fprintf(stderr, "Unknown argument %s\n", argv[i]); return 1;
			}
		}
	}

    UINT32 u32 = *u32Runs;
    *u32Runs += 3;
    (*u32Runs) &= (*u32Runs) & (0xFFFFFFFF - 3);
    if( (*u32Runs)  != u32 )
    {
        printf("\n Warning! runs must be divisible by 4. Please change runs (-r) to %u \n ", *u32Runs);
    }
    if(((*u32Dim) * (*u32Runs)) > 7000)
    {
        printf("\n Error! dimension x runs must not larger than 7000 because of local store limitation-> exit \n");
	return 1;
    }
    if(((*u32Dim) * (*u32Runs) * (*u32Iter)) > 6250000)
    {
        printf("\n Error! dimension x runs x iteration  must not larger than 6 250 000 because of main memory limitation-> exit \n");
        return 1;
    }
    if( (*u32Dim < 1) || (*u32Dim > 40) )
    {
        printf("\n Error! Dimension is %u. It must be in the range of 1 .. 40. -> exit \n", *u32Dim);
        return 1;
    }

    return 0;
}

// **************************************************

// (1) spe handles and variable to keep SPE status
extern spe_program_handle_t spu_rand;
extern spe_program_handle_t spu_MC;   

// (2) static allocation of memory aligned to 128 byte boundary
spu_rand_CB_t  spu_randCB[NUMSPUS]  __attribute__ ((aligned(128)));	// control block


typedef struct thread_args {
	spe_context_ptr_t spe;
	void * argp;
	void * envp;
} thread_args_t;

void *spe_thread(void *arg){
	unsigned int runflags = 0;
	unsigned int entry = SPE_DEFAULT_ENTRY;
	thread_args_t * targ = (thread_args_t *)arg;
	// run SPE context
	if(spe_context_run(targ->spe, &entry, runflags, targ->argp, targ->envp, NULL) <0 ) {
		printf("** spe_context_run error, errno = %d\n", errno);
		exit(10);
	}
	pthread_exit(NULL);
}


// ***********************************************************************************************************
int main(int argc, char* argv[])
{
        UINT32 rng_cycles[NUMSPUS];//, mc_cycles[NUMSPUS];	
	UINT8  u8Report = 0;
	UINT32 u32Dim, u32Runs, u32Iter;
	UINT32 i, u32BufferLen;
	vector float * afRandBuffer[NUMSPUS];
	char strRNG[8][36];

	if( 0 != get_args(argc, argv, &u32Dim, &u32Runs, &u32Iter, &u8Report) )
		return 1;

	if(u8Report != 0)
		printf("\n n-dim sphere: example for SDK RNG");


	u32BufferLen = (u32Iter * u32Dim * u32Runs * sizeof(vector float) + 127) & 0xFFFFFF80;

        pthread_t pts[NUMSPUS];
	spe_context_ptr_t spe[NUMSPUS];
	thread_args_t targ1[NUMSPUS];
	int rc;


	// create different random numbers on different SPEs in parallel
	for ( i = 0; i < NUMSPUS; i++)
	{
	    // (3) dynamic allocation of memory aligned to 128 = 2^7 byte boundary
	    afRandBuffer[i] = (vector float *)malloc_align(u32BufferLen, 7);	

	    if (afRandBuffer[i] == NULL)
	    {
	        printf("\n not enough memory: requested was %u", u32BufferLen);
	        return 0;
	    }
	

		// >> set up control block (4)
		spu_randCB[i].pfRandBufferMainMem = afRandBuffer[i];
		spu_randCB[i].u32Len  = u32BufferLen;
		spu_randCB[i].u32Iter = u32Iter;
		spu_randCB[i].u32Dim  = u32Dim;
		spu_randCB[i].u32Runs = u32Runs;
		spu_randCB[i].u32Index = i;
		spu_randCB[i].u32Cycles = 0;
		strcpy(spu_randCB[i].strRNG, strRNG[i]);
		rand_sobol_set_CB( &(spu_randCB[i].sobolCB), 0);
		// <<

		// create thread  (5)
		if( (spe[i] = spe_context_create(0, NULL)) == NULL) {
			printf("** spe_context_create error, i=%d, errno = %d\n", i, errno);
			exit(2);
		}

		if( spe_program_load(spe[i], &spu_rand)!=0) {
			printf("** spe_program_load error, i=%d, errno = %d\n", i, errno);
			exit(3);
		}
   
		targ1[i].spe = spe[i];
		targ1[i].argp = (void *)&spu_randCB[i];
		targ1[i].envp = NULL;

		// Create thread
		if ((rc = pthread_create(&pts[i], NULL, spe_thread, (void *)&targ1[i])) != 0) {
			printf ("** pthread_create error, error=%d\n", rc);
			exit(6);
		}



	}

        INT32 spe_done[2];
	INT32 count = 0;
	INT32 temp = 0;

        while ( count < NUMSPUS)
	{
		for(i=0;i< NUMSPUS;i++)  // pull mail box data from each SPE and handle message
		{ 
			// get mail box message from SPEs  (6)
                        spe_done[0] = 0;
			if((temp = spe_out_mbox_read(spe[i],(UINT32*)&spe_done[0], 1 ) > 0)) {
//				printf("spe(%d): spe_done[0] = %d  temp = %d\n", i, spe_done[0], temp);
			}

			if(spe_done[0] == 1)  // check if  "SPE executing spu_rand" is done
			{
				// wait until SPE is idle and create new SPE thread on idle SPE (7)
				pthread_join(pts[i], NULL);

                                rng_cycles[i] =  spu_randCB[i].u32Cycles;		// save cycle counts from RNG
		                strcpy(strRNG[i], spu_randCB[i].strRNG);

                
				// Destroy context
				if ((rc = spe_context_destroy (spe[i])) != 0) {
					printf ("**spe_context_destroy error, i=%d, errno=%d \n", i, rc);
					exit (20);
				}                

				if( (spe[i] = spe_context_create(0, NULL)) == NULL) {
					printf("** spe_context_create error, i=%d, errno = %d\n", i, errno);
					exit(2);
				}

				if( spe_program_load(spe[i], &spu_MC)!=0) {
					printf("** spe_program_load error, i=%d, errno = %d\n", i, errno);
					exit(3);
				}
   
				targ1[i].spe = spe[i];
				targ1[i].argp = (void *)&spu_randCB[i];
				targ1[i].envp = NULL;

				// Create thread
				if ((rc = pthread_create(&pts[i], NULL, spe_thread, (void *)&targ1[i])) != 0) {
					printf ("** pthread_create error, error=%d\n", rc);
					exit(6);
				}

			}
			else if(spe_done[0] == 2) {  // check if "SPE executing spu_MC" is done
				count++;
			}
		}
	}
	
	// >> print the results
	u8Report &= 0x7F;
        for ( i = 0; i < NUMSPUS; i++)
	{

	    printf("\n");
	    if ( u8Report == 1)
            {
		printf("\n dimension = %u", spu_randCB[0].u32Dim);
            	printf(" iteration = %u, runs = %u", spu_randCB[0].u32Iter, spu_randCB[0].u32Runs);

 		printf("\n spu-call %i: RNG = %s used %u decr_ticks",i, strRNG[i], rng_cycles[i]);
                printf("\n spu-call %i: RNG = %s used %u decr_ticks",i, spu_randCB[i].strRNG, spu_randCB[i].u32Cycles);
                printf("\n %u-dimensional sphere with radius %f has volume of %e \n",
                        spu_randCB[i].u32Dim, spu_randCB[i].radius, spu_randCB[i].volume);
                printf(" iteration = %u, runs = %u", spu_randCB[i].u32Iter, spu_randCB[i].u32Runs);
//		printf("\tmean=%e, sigma=%e, delta=%e", spu_randCB[i].mean, spu_randCB[i].sigma, spu_randCB[i].delta);
            }
	    if ( u8Report == 2 )
               printf("%u, ticks: %-15s = %7u %-36s = %7u", i, strRNG[i], rng_cycles[i], spu_randCB[i].strRNG, spu_randCB[i].u32Cycles);

            printf("\tmean = %e, sigma = %e, delta = %e", spu_randCB[i].mean, spu_randCB[i].sigma, spu_randCB[i].delta);
        }
            printf("\n");


	printf("\n");
	return 0;
}
