/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/******************************************************************************
 * spu_MC program 
 *
 * This SPE program perofrmas the Monte Carlo simulation for the PPE sphereMC_ppu.c program.
 * All control parameters are transfered to this SPE task by control block.
 * The SPE loads the necessary random numbers from main memory into its local store in sub-blocks.
 *
 *
 * The PPU example from sphereMC_ppu.c describes:  
 * - receiving information from PPE by control block
 * - setting DMA - mask
 * - wait until all DMA transfers, where the mask is set, are done
 * - profile with decrementer
 * - handle DMA transfers that are larger than 4kByte
 * - get number of emty slots in DMA queue
 * - prevent an overflow of the DMA queue
 * - store data (< 4kB) from local store into main memory
 *
 *
 * This SPU example demonstrates how to:
 * - perform math on vector (1)
 * - replace an if - condition by a spu_compare intrinsic (2) 
 * - use a vector as multiple counters (3)
 * - condense information of vector into a scalar in O(log(n)) operations (4)
 * - dynamic branch prediction (5)
 * - perform double buffering (6)
 * - perform DMA transfers on quadword (128 byte) aligned (6) 
 *  (ref. Cell Broadband Engine Programming Handbook 19.2.1)
 *
 *****************************************************************************/


#include <stdio.h>
#include <spu_mfcio.h>
#include <libmisc.h>
#include <math.h>
#include <string.h>

// Note: These three are order dependent to ensure the Sobol control block and local types are defined.
#include <mc_rand_sb.h>
#include "../myType.h"
#include "mc_sphere.h"


#define RADIUS (float)0.5
#define RADIUS_SQUARE (float)RADIUS * (float)RADIUS

#define DMA_TAG_0 0
#define DMA_TAG_1 1


/***************************************************************
 * Module to to the Monte Carlo simulation
 * modification to scalar code:
 *
 * scalar vers SIMD  math:
   scalar           operations    SIMD                                         operations
   x = 2.0*x -1.0;    2	       -> x = spu_msub(float_vec_2, x, float_vec_1);   1 
   x = x * f_radius;  1        -> x = spu_mul(x, float_vecSphereR);            1 
   r += x*x;          2        -> r = spu_madd(x, x, r);                       1
 * result
 * beside the fact that SIMD operation operates on 4 elements in this case, 
 * the number of operations is reduced from 5 (scalar) to 3 (SIMD) 
 * performance increase ~ 6 - 7 times
 *
 *
 * Hit test, scalar vers SIMD
   scalar                   operations    SIMD                                                          operations
   N++;                     1             N+=4                                                             1
   if((r<=RADIUS_SQUARE))   1             u32_vecHitResult = spu_cmpabsgt(float_vecSphereRsquare, r);      1
       ++nHit;              1             u32_vecHitResult = spu_and(u32_vecHitResult, u32_vecAdd);        1
                                          u32_vec_nHit = spu_add(u32_vec_nHit, u32_vecHitResult);          1
 * result
 * beside the fact that SIMD operation operates on 4 elements in this case,
 * the number of operations is increaed from 3 (scalar) to 4 (SIMD)
 * performance increase ~ 3 times

 * summary
   performance increase: 4 x (5 x dim + 3) / (3 x dim +4)
   dim = 1:        4 x 8 / 7      ~ 4
   dim = 40:       4 x 203 / 124  ~ 6.5
   lim (dim -> oo) 4 x 5 / 3      ~ 6.7 

*************************************************************************************************/
#define vec(x,i) spu_extract(x,i)
#define print_f4(x) printf(" %f %f %f %f", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_u4(x) printf(" %u %u %u %u", vec(x,0), vec(x,1), vec(x,2), vec(x,3))
#define print_x4(x) printf(" %x %x %x %x", vec(x,0), vec(x,1), vec(x,2), vec(x,3))

// ***************************************************************************
UINT8 checkParameter(spu_rand_CB_t * pcb)
{
        UINT32 u32Len, u32Addr, u32Flag;

        u32Flag = 0;

        // check if array in main memory is large enough
        u32Len = ((pcb->u32Iter * pcb->u32Dim * pcb->u32Runs * sizeof(vector float)) + 127)&0xFFFFFF80;
        if ( u32Len > pcb->u32Len)
        {
                printf("\n Program terminated: not enough memory in main memory\n");
                u32Flag |= 1;
        }
        // check if memory area in main memory is 128 byte aligned
        u32Addr = (UINT32)pcb->pfRandBufferMainMem;
        if( u32Addr != (u32Addr & 0xFFFFFF80) )
        {
                printf("\n Program terminated: main memory not 128 byte aligned\n");
                u32Flag |= 2;
        }
        return u32Flag;
}
// ************************************************************************
double get_volume_sphere( double radius, UINT32 dimension)
{
        //  calculate the exact volume of n-dim sphere to compare it with the Monte Carlo result
        double pi_val, volume0, volume1, volume;

        pi_val = (double)3.1415926535897932384626433832795;
        volume0 = (double)1.0;
        volume1 = (double)2.0 * radius;
        if (dimension % 2 == 0)
                volume = volume0;
        else
                volume = volume1;
        UINT32 u;
        for (u = 1; u < dimension; u++)
        {               volume = volume0 * (double)2.0 / (double)(u+1) * pi_val * radius * radius;
                        volume0 = volume1;
                        volume1 = volume;
        }
	return volume;
}
//  ************************************************************

// static allocation of aligned memory (128 byte alligned)
spu_rand_CB_t cb __attribute__ ((aligned(128)));                // local copy of control block in main memory

int main(UINT64 speid, UINT64 argp/*, UINT64 envp*/)
{
	speid = 0;

	UINT32 u32Error;

        // get the contents of the control block from main memory to local store
        mfc_get(&cb, (UINT32)argp,sizeof(spu_rand_CB_t),DMA_TAG_0,0,0);
        mfc_write_tag_mask(1<<DMA_TAG_0);
        mfc_read_tag_status_all();

	UINT32 u32_ts, u32DecrCounts;

	double (*ptr2aFunc[8])(UINT32, float , UINT32, vector float*, spu_rand_CB_t*);
	ptr2aFunc[0] =  &sphereIntegration_0;
        ptr2aFunc[1] =  &sphereIntegration_1;
        ptr2aFunc[2] =  &sphereIntegration_2;
        ptr2aFunc[3] =  &sphereIntegration_3;
        ptr2aFunc[4] =  &sphereIntegration_4;
        ptr2aFunc[5] =  &sphereIntegration_4;
        ptr2aFunc[6] =  &sphereIntegration_4;
        ptr2aFunc[7] =  &sphereIntegration_4;


	double (*ptr2Func)(UINT32, float , UINT32, vector float*, spu_rand_CB_t*);
	ptr2Func = ptr2aFunc[(cb.u32Index & 0x7)];

//--------------------------------------------------------------------------------------------------------		
	//  calculate the exact volume of n-dim sphere to compare it with the Monte Carlo result
	double volume = get_volume_sphere((double)RADIUS , cb.u32Dim);


	// variables needed for the Monte Carlo simulation
        UINT32 u, u32Runs, u32Iter, u32Dim;
        double sum, sumR, ssumR;

	u32Runs = cb.u32Runs;
	u32Dim = cb.u32Dim;
	u32Iter = cb.u32Iter;
	sumR = ssumR  = 0;
// ---------------------------------------------------------------------------------

        // variables and intitialisation for double buffering
        // transfer size = size of one iteration + max(misalignment between local and main memory) (6)
	const UINT32 u32BlockSize = u32Dim * u32Runs*sizeof(vector float);             // bytes per iteration used
	const UINT32 u32TransferSize = ((u32BlockSize + 127)&(0xFFFFFFFF - 127))+128 ; // bytes per iteration transfered
        const UINT32 u32MaxDMA = (MAX_DMA & (0xFFFFFFFF - 127));  // max block size must be multiple of 128 (6)
        UINT32 au32Tag[2], u32Flag0, u32Flag1;
	UINT32 u32SourceMainMem, u32Transfer, u32Head, u32Lost[2];
	vector float * pfCacheLine_array[2], *pf_vec[2];

        u32Flag0 = 0;
        u32Flag1 = 1;
        au32Tag[0] = 0;
        au32Tag[1] = 2;

	u32SourceMainMem   = (UINT32)cb.pfRandBufferMainMem;  // pointer to first RN in main memory

	// check if there are enough RNS in main memory 
        u32Error = checkParameter(&cb);
        if (0 != u32Error)
                return (int)u32Error;

       // begin of double buffered DMA transfer (6)


	// two cache lines to hold actual and next data, 128 byte aligned (6)
	pfCacheLine_array[0] = (vector float *)malloc_align(u32TransferSize,7);
	pfCacheLine_array[1] = (vector float *)malloc_align(u32TransferSize,7); 

	// start DMA Transfer for first data block
        u32Transfer = u32TransferSize ;
        pf_vec[u32Flag0] = pfCacheLine_array[u32Flag0];

	u32Head = u32SourceMainMem;
        u32Head = u32SourceMainMem & (0xFFFFFFFF - 127);      // get next 128 byte aligned address less than source address (6)
        u32Lost[u32Flag0] = u32SourceMainMem - u32Head;                 // calculate gap between head and source address (6)
	
        while( u32Transfer > u32MaxDMA)
        {
            if( mfc_stat_cmd_queue() > 2) // check if DMA has enough slots
            {
                mfc_get(pf_vec[u32Flag0], u32Head, u32MaxDMA, au32Tag[u32Flag0], 0, 0);
                u32Transfer -= u32MaxDMA;
                u32Head += u32MaxDMA;
                pf_vec[u32Flag0] = (vector float *)((char *)pf_vec[u32Flag0] + u32MaxDMA);
            }
        }
        mfc_get(pf_vec[u32Flag0], u32Head, u32Transfer, au32Tag[u32Flag0], 0, 0);
	u32SourceMainMem += u32BlockSize;
        u32DecrCounts = 0;

	// continue DMA Transfer and do Monte Carlo Simulation in parallel
       	for( u = 1; u < u32Iter; u++)
       	{
               mfc_write_tag_mask(1<<au32Tag[u32Flag0]); // set Tag for data block used in this itaration

		// start DMA Transfer for next data block
		u32Transfer = u32TransferSize;
		pf_vec[u32Flag1] = pfCacheLine_array[u32Flag1];

		u32Head = u32SourceMainMem;
	        u32Head = u32SourceMainMem & (0xFFFFFFFF - 127);      // get next 128 byte aligned address less than source address (6)
	        u32Lost[u32Flag1] = u32SourceMainMem - u32Head; 

		while( u32Transfer >  u32MaxDMA)
		{
                    if( mfc_stat_cmd_queue() > 2)
                    {
                        mfc_get(pf_vec[u32Flag1], u32Head, u32MaxDMA, au32Tag[u32Flag1], 0, 0);
                        u32Transfer -= u32MaxDMA;
                        u32Head += u32MaxDMA;
  	 	        pf_vec[u32Flag1] = (vector float *)((char *)pf_vec[u32Flag1] + u32MaxDMA);
                    }		
		}
        	mfc_get(pf_vec[u32Flag1], u32Head, u32Transfer, au32Tag[u32Flag1], 0, 0);
	        u32SourceMainMem += u32BlockSize;

		mfc_read_tag_status_all();		  // wait until data block for this interation is in local store

		// call your Monte Carlo simulation function, skip lost bytes (6)
                spu_write_decrementer(-1); 
                u32_ts = spu_read_decrementer(); 
		sum = ptr2Func(u32Runs, RADIUS ,u32Dim, (vector float*)((char *)pfCacheLine_array[u32Flag0] + u32Lost[0]), &cb);
                u32_ts -= spu_read_decrementer(); 
                u32DecrCounts += u32_ts;
		// collect the MC results
                sumR  += sum;
                ssumR += sum*sum;
		//  toggle flags => next data block  will be this data block
                u32Flag0 ^= 0x1;
                u32Flag1 ^= 0x1;
	}
	// post process double buffering; execute last loaded cache line
        mfc_write_tag_mask(1<<au32Tag[u32Flag0]); //  set Tag for data block used in the last simulation
        mfc_read_tag_status_all();                // wait for data block

        spu_write_decrementer(-1); 
        u32_ts = spu_read_decrementer();
        sum = ptr2Func(u32Runs, RADIUS ,u32Dim, (vector float*)((char *)pfCacheLine_array[u32Flag0] + u32Lost[0]), &cb);
        u32_ts -= spu_read_decrementer();
        u32DecrCounts += u32_ts; 

        // collect results on Monte Carlo simulations
        sumR  += sum;
        ssumR += sum*sum;

//-------------------------------------------------------------------------------- end of double buffered DMA transfer
	// summarize the results
	double scale, sscale;
	scale = pow((double)RADIUS*(double)2.0, u32Dim) / (double)u32Iter;	// scaling factor for mean
	sscale = scale * pow((double)RADIUS*(double)2.0, u32Dim);			// scaling factor for varianz
	
	sumR = sumR * scale;			// calculate mean
        ssumR = ssumR *sscale;			
        ssumR = sqrt(ssumR - (sumR * sumR));	// calculate sigma
	ssumR = ssumR / volume * 100.0;         // scale sigma in % of true value


	// write results to control block
	cb.mean = (float)sumR;
	cb.sigma = (float)ssumR;
	cb.radius = (float)RADIUS;
	cb.volume = (float)volume;
	cb.delta = (float)((double)sumR/ (double)volume *(double)100 - 100);   // scale delta in % of true value

	//
        cb.u32Cycles = u32DecrCounts;

	// put results to main memory
        mfc_put(&cb, (UINT32)argp, 128, DMA_TAG_1, 0, 0);
        mfc_write_tag_mask(1<<DMA_TAG_1);
        mfc_read_tag_status_all();


        spu_write_out_mbox(2);   // notify PPU that volume of sphere results are available 

	return 0;
}
