/* --------------------------------------------------------------  */
/* (C)Copyright 2007                                               */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */

/******************************************************************************
 * spu_rand.c: 
 *
 * This SPE programm creates random numbers for the PPE sphereMC_ppu.c program.
 * All control parameters are transfered to this SPE task by control block.
 * The SPE creates the random numbers in block of length runs and does this iter (iteration)
 * times.
 *
 * The PPU example from sphereMC_ppu.c describes: 
 * - how to set up a SPE task
 * - how to do static aligned memory allocation
 * - how to do dynamic aligned memory allocation
 * - how to set up a SPE control block
 * - how to create spe thread
 * - how to use mail boxes
 * - how to restart a SPEs imediatly with a new task
 *
 *
 *
 * This SPU example demonstrates how to
 * - get information from PPE by control block (1)
 * - set DMA - mask (2)
 * - wait until all DMA transfers, where the mask is set, are done (3)
 * - profile with decrementer (4)
 * - handle DMA transfers that are larger than 4kByte (5)
 * - get number of empty slots in DMA queue (7)
 * - prevent an overflow of the DMA queue (6)
 * - store data (< 4kB) from local store into main memory (8) 
 *
 *****************************************************************************/

#include <stdio.h>
#include <spu_mfcio.h>
#include <string.h>
#include <libmisc.h>
#include <rand_0_to_1_v.h>

// NOTE: These includes are order sensitive
#include <mc_rand.h>
#include <../myType.h>
#include "spu_rand.h"


#define DMA_TAG_0 0
#define DMA_TAG_1 1

// ***************************************************************************
UINT8 checkParameter(spu_rand_CB_t * pcb)
{
	UINT32 u32Len, u32Addr, u32Flag;

	u32Flag = 0;

        // check if array in main memory is large enough
        u32Len = ((pcb->u32Iter * pcb->u32Dim * pcb->u32Runs * sizeof(vector float)) + 127)&0xFFFFFF80; 
        if ( u32Len > pcb->u32Len)
        {
                printf("\n Program terminated: not enough memory in main memory\n");
                u32Flag |= 1;
        }
	// check if memory area in main memory is 128 byte aligned
	u32Addr = (UINT32)pcb->pfRandBufferMainMem;
	if( u32Addr != (u32Addr & 0xFFFFFF80) )
	{
		printf("\n Program terminated: main memory not 128 byte aligned\n");
		u32Flag |= 2;
	}
	return u32Flag;
}
// ************************************************************************

spu_rand_CB_t cb __attribute__ ((aligned(128)));		// locale copy of control block in main memory

int main(UINT64 speid, UINT64 argp/*, UINT64 envp*/)
{

        speid = 0;
        // get the contents of the control block from main memory to local store (1)
        // control block is defintly smaller that 4kByte, hence it can be loaded by one DMA task 
        mfc_get(&cb, (UINT32)argp,/* sizeof(spu_rand_CB_t)*/ 128, (UINT32)DMA_TAG_0, 0, 0);
        // set DMA - TAG mask to control DMA status (2)
        mfc_write_tag_mask(1<<DMA_TAG_0);
        // wait until all pending DMA transfers are done (3)
        mfc_read_tag_status_all();

	// define function pointer to different RNGs
        void (*ptr2aFunc[8])(UINT32, UINT32, vector float **, char *, UINT32 *);
        ptr2aFunc[0] =  &scalar_0_to_1_array_f4;
        ptr2aFunc[1] =  &sdk_0_to_1_array_f4;
        ptr2aFunc[2] =  &mt_0_to_1_array_f4;
        ptr2aFunc[3] =  &ks_0_to_1_array_f4;
        ptr2aFunc[4] =  &sb_0_to_1_array_f4;
	ptr2aFunc[5] =  &hw_0_to_1_array_f4;
        ptr2aFunc[6] =  &ks_0_to_1_array_f4;
        ptr2aFunc[7] =  &sb_0_to_1_array_f4;

	//  choose fucntion
        void (*ptr2Func)(UINT32, UINT32, vector float**, char*, UINT32 *);
        ptr2Func = ptr2aFunc[(cb.u32Index & 0x7)];

        // check parameter especially if there is enough memory in main memory
        UINT32 u32Error = checkParameter(&cb);
        if (0 != u32Error)
                return (int)u32Error;


	// parameters for DMA transfer of data (RNs)
	const UINT32 u32TransferSize = cb.u32Dim * cb.u32Runs * sizeof(vector float);

	// parameter for RNG	        
	const UINT32 u32Runs = cb.u32Runs;
	const UINT32 u32Dim = cb.u32Dim;

	UINT32 i;

	UINT32 u32DestMainMem, u32Transfer;
	UINT32 rng_ts=0, u32DecrCounts;
        vector float * pf_vec, * pf_vecRand, * pf_vecCacheLine;

	// check parameter especially if there is enough memory in main memory
	u32Error = checkParameter(&cb);
	if (0 != u32Error)
		return (int)u32Error;

	UINT32 u32CacheSize = ((((u32Runs + 76)*u32Dim) + 8) * sizeof(vector float) + 127)&0x0FFFFFF80;
	u32DestMainMem = (UINT32)cb.pfRandBufferMainMem;

	pf_vecCacheLine = (vector float *)malloc_align(u32CacheSize,7);   // dynamic allocation of 128 = 2^7 byte aligned memory
	pf_vecRand = pf_vecCacheLine;


        if((cb.u32Index & 0x7) == 5)  // hw rng is to be used in this spe
        {
            if(mc_rand_hw_init() <0)// running on System Simulator
            {
                printf("\n *** This environment does not support the Hardware Random Number Generator \n");
            }
        }
	
        mc_rand_ks_init(1);  // init KS
        mc_rand_mt_init(0);  // init MT

        // init SB
        vector unsigned char *rand_sb_mem = (vector unsigned char *)((char * )pf_vecCacheLine);
        mc_rand_sb_init(&(cb.sobolCB), u32Runs, cb.u32Dim, rand_sb_mem, u32CacheSize );
        mc_rand_sb_seed(0);


	// >> create random numbers
	u32DecrCounts = 0;
	for(i=0; i < cb.u32Iter; i++)
	{
		ptr2Func(u32Runs, u32Dim, &pf_vecRand, cb.strRNG, &rng_ts);  // run RNG
		u32DecrCounts += rng_ts;  // sum up all decrementer counts (4)
       	
	        // handle data transfer to main memory if the data sieze could be larger than 4kB (5)
                u32Transfer = u32TransferSize;
		pf_vec = pf_vecRand;
		UINT32 u32DMAqueue;
		while (u32Transfer > MAX_DMA)
		{
		    // get numbers of empty slots in DMA queue (6)
		    u32DMAqueue = mfc_stat_cmd_queue(); 
		    if( u32DMAqueue > 1)  // we need to keep one free slot for the DMA transfer following this loop see (7.1)
		    {	
			// initiate DMA transfer of a block of maximum size (MAX_DMA), update the pointer to local
			// store, main memory and the number of bytes to be transfered  (7)
                        mfc_put(pf_vec,  u32DestMainMem, MAX_DMA, DMA_TAG_1,0,0);
		        pf_vec = (vector float *)( (char *)pf_vec + MAX_DMA);
		        u32DestMainMem  += MAX_DMA; 
    		        u32Transfer -= MAX_DMA;
                    }
	        }
                // initiate the last DMA transfer (7.1)
                mfc_put(pf_vec,  u32DestMainMem, u32Transfer, DMA_TAG_1,0,0); 
                mfc_write_tag_mask(1<<DMA_TAG_1);
	        u32DestMainMem  += u32Transfer;   // update pointer to main memory
                mfc_read_tag_status_all();

        }

        //  put results to main memory, if data size is less 4kB only (8)
        cb.u32Cycles = u32DecrCounts;          

        mfc_put(&cb, (UINT32)argp, sizeof(spu_rand_CB_t), DMA_TAG_1, 0, 0);
        mfc_write_tag_mask(1<<DMA_TAG_1); //set mask like in (2)
        mfc_read_tag_status_all();  // wait untill DMA transfer is done like (3)

        spu_write_out_mbox(1); // notify PPU that random number data is available     

	return 0;
}
